<template>
    <div class="model-container" ref="model-container">
        <div class="view-container" ref="threeDBox"></div>
        <div id="loading">
            <el-progress
                type="circle"
                :percentage="percent"
                :show-text="false"
                :stroke-width="14"
            ></el-progress>
        </div>
    </div>
</template>

<script>
import * as THREE from 'three'
import {OrbitControls} from 'three/examples/jsm/controls/OrbitControls.js'
// import {FBXLoader} from 'three/examples/jsm/loaders/FBXLoader.js'
import {GLTFLoader} from 'three/examples/jsm/loaders/GLTFLoader.js'

import {GUI} from 'three/examples/jsm/libs/lil-gui.module.min.js'
import moment from 'moment'

// 用于模型边缘高亮
import {EffectComposer} from 'three/examples/jsm/postprocessing/EffectComposer.js'
import {RenderPass} from 'three/examples/jsm/postprocessing/RenderPass.js'
import {OutlinePass} from 'three/examples/jsm/postprocessing/OutlinePass.js'
import {OutputPass} from 'three/examples/jsm/postprocessing/OutputPass.js'
import {ShaderPass} from 'three/examples/jsm/postprocessing/ShaderPass.js'
import {FXAAShader} from 'three/examples/jsm/shaders/FXAAShader.js'
import {SMAAPass} from 'three/examples/jsm/postprocessing/SMAAPass.js'
import {UnrealBloomPass} from 'three/examples/jsm/postprocessing/UnrealBloomPass.js'

export default {
    name: 'ModelIndex',
    data() {
        return {
            width: 0,
            height: 0,
            scene: null,
            camera: null,
            controls: null,
            renderer: null,
            css2Renderer: null,
            clock: null,
            mixer: null,
            stats: null,
            animationId: null,
            chooseObj: null,
            tag: null,
            percent: 0,
            gui: null,
            composer: null, //效果组合器
            outlinePass: null, //发光通道
            effectFXAA: null, //自定义的着色器通道
        }
    },
    mounted() {
        this.$nextTick(() => {
            this.width = this.$refs['model-container'].offsetWidth
            this.height = this.$refs['model-container'].offsetHeight
            this.init()

            this.animate()
        })
    },
    methods: {
        addComposer() {
            // 创建一个EffectComposer（效果组合器）对象，然后在该对象上添加后期处理通道。
            this.composer = new EffectComposer(this.renderer)
            // 新建一个场景通道  为了覆盖到原来的场景上
            let renderPass = new RenderPass(this.scene, this.camera)
            this.composer.addPass(renderPass)
            // 物体边缘发光通道
            this.outlinePass = new OutlinePass(
                new THREE.Vector2(this.width, this.height),
                this.scene,
                this.camera
            )
            this.outlinePass.visibleEdgeColor.set(parseInt(0x00ff00)) // 呼吸显示的颜色
            this.outlinePass.hiddenEdgeColor = new THREE.Color(0, 0, 0) // 呼吸消失的颜色
            this.composer.addPass(this.outlinePass)

            // 解决高亮后环境变暗的问题
            const outputPass = new OutputPass()
            this.composer.addPass(outputPass)

            // 自定义的着色器通道 作为参数
            this.effectFXAA = new ShaderPass(FXAAShader)
            this.effectFXAA.uniforms['resolution'].value.set(
                1 / this.width,
                1 / this.height
            )
            this.composer.addPass(this.effectFXAA)
        },
        loadModel() {
            const loader = new GLTFLoader()
            loader.load(
                '/static/3DModel/sfz-lh.gltf',
                (gltf) => {
                    this.percent = 100

                    this.scene.add(gltf.scene)
                    document.getElementById('loading').style.display = 'none'
                },
                (xhr) => {
                    let percent = Math.floor((xhr.loaded / xhr.total) * 100)
                    if (percent < 96) {
                        this.percent = percent
                    }
                },
                (error) => {
                    console.log(error)
                }
            )
        },
        GUIControler(mesh) {
            this.gui = new GUI()

            this.gui.domElement.style.right = '0px'
            this.gui.domElement.style.top = '100px'
            this.gui.domElement.style.width = '300px'

            this.gui.add(mesh.position, 'x', -100, 100).name('X坐标').step(0.1)
            this.gui.add(mesh.position, 'y', -100, 100).name('Y坐标').step(0.1)
            this.gui.add(mesh.position, 'z', -100, 100).name('Z坐标').step(0.1)
        },
        init() {
            let container = this.$refs.threeDBox
            console.log(container, 'container')
            this.scene = new THREE.Scene()

            // 创建纹理加载器
            const textureLoader = new THREE.TextureLoader()
            // 创建天空球
            const skyGeometry = new THREE.SphereGeometry(1000, 60, 60)
            const skyTexture = textureLoader.load(
                // './static/3DModel/img/sky.jpg'
                '/static/3DModel/img/sky.png'
            ) // 天空纹理
            const skyMaterial = new THREE.MeshBasicMaterial({
                map: skyTexture, // 颜色贴图
            })
            // 反转球体，主要是反转z轴
            skyGeometry.scale(1, 1, -1)
            const sky = new THREE.Mesh(skyGeometry, skyMaterial)
            this.scene.add(sky)

            this.camera = new THREE.PerspectiveCamera(
                30,
                this.width / this.height,
                1,
                3000
            )
            this.camera.position.set(-82.3681, 4.6218, 0.3505)
            this.camera.lookAt(0, 0, 0)

            this.loadModel() //加载模型

            // const axesHelper = new THREE.AxesHelper(100)
            // this.scene.add(axesHelper)

            const ambient = new THREE.AmbientLight(0xffffff, 1)
            this.scene.add(ambient)

            const directionalLight = new THREE.DirectionalLight(0xffffff, 2)
            // 设置光源的方向：通过光源position属性和目标指向对象的position属性计算
            directionalLight.position.set(80, 100, 50)
            // 方向光指向对象网格模型mesh，可以不设置，默认的位置是0,0,0
            // directionalLight.target = mesh
            this.scene.add(directionalLight)

            const cubeMaterial = [
                //side
                new THREE.MeshBasicMaterial({
                    color: 0xffffff, //设置材质颜色
                    transparent: true, //开启透明
                    opacity: 1, //设置透明度
                    side: THREE.FrontSide,
                }),
                //top
                new THREE.MeshBasicMaterial({
                    map: new THREE.TextureLoader().load(
                        '/static/3DModel/img/video.jpg'
                    ),
                    side: THREE.DoubleSide,
                }),
                //bottom
                new THREE.MeshBasicMaterial({
                    map: new THREE.TextureLoader().load(
                        '/static/3DModel/img/video.jpg'
                    ),
                    side: THREE.DoubleSide,
                }),
            ]
            const geometry = new THREE.CylinderGeometry(0.15, 0.15, 0.05)
            // geometry.rotateX(Math.PI / 2)
            // geometry.rotateZ(Math.PI / 2)
            // console.log('rotate', geometry)
            const mesh = new THREE.Mesh(geometry, cubeMaterial)
            mesh.position.set(8.2, 6.9, 1.3)
            mesh.rotation.set(0, 0, Math.PI / 2, 'XZY')
            this.scene.add(mesh) //网格模型添加到场景中

            this.GUIControler(mesh)

            this.renderer = new THREE.WebGLRenderer({antialias: true})
            this.renderer.setPixelRatio(window.devicePixelRatio)
            this.renderer.setSize(this.width, this.height)

            this.renderer.outputColorSpace = THREE.SRGBColorSpace //设置为SRGB颜色空间
            container.appendChild(this.renderer.domElement)

            this.controls = new OrbitControls(
                this.camera,
                this.renderer.domElement
            )
            // -65.1083, 4.2508, 0.851
            this.controls.target = new THREE.Vector3()
            this.controls.minDistance = 0
            this.controls.maxDistance = 200
            this.autoRotate = true
            this.controls.update()
            window.addEventListener('resize', this.onWindowResize)

            this.renderer.domElement.addEventListener('click', (event) => {
                // .offsetY、.offsetX以canvas画布左上角为坐标原点,单位px
                const px = event.offsetX
                const py = event.offsetY
                // 屏幕坐标px、py转WebGL标准设备坐标x、y
                // width、height表示canvas画布宽高度
                const x = (px / this.width) * 2 - 1
                const y = -(py / this.height) * 2 + 1

                //创建一个射线投射器`Raycaster`
                const raycaster = new THREE.Raycaster()

                //.setFromCamera()计算射线投射器`Raycaster`的射线属性.ray
                // 形象点说就是在点击位置创建一条射线，射线穿过的模型代表选中
                raycaster.setFromCamera(new THREE.Vector2(x, y), this.camera)
                //.intersectObjects([mesh1, mesh2, mesh3])对参数中的网格模型对象进行射线交叉计算
                // 未选中对象返回空数组[],选中一个对象，数组1个元素，选中两个对象，数组两个元素
                // const intersects = raycaster.intersectObjects([mesh])
                const intersects = raycaster.intersectObject(this.scene, true)
                console.log('射线器返回的对象', intersects)
                // intersects.length大于0说明，说明选中了模型
                if (intersects.length > 0) {
                    // 选中模型的第一个模型，设置为红色
                    // intersects[0].object.material.color.set(0xff0000)
                    // intersects[0].object.add(this.tag)
                    this.drop()
                    this.chooseObj = intersects[0].object
                    this.pick(this.chooseObj)
                } else {
                    this.drop()
                    // mesh.material.color.set(0xfff000)
                    // if (this.chooseObj) {
                    //     //把原来选中模型对应的标签和发光描边隐藏
                    //     this.chooseObj.remove(tag) //从场景移除
                    // }
                }
            })
            this.addComposer()
            setTimeout(() => {
                this.state = '异常'
            }, 10000)
        },
        drop() {
            this.outlinePass.selectedObjects = []
        },
        pick(selectedObject) {
            this.outlinePass.selectedObjects = [selectedObject]
        },
        onWindowResize() {
            this.camera.aspect = this.width / this.height
            this.renderer.setSize(this.width, this.height)
            this.composer.setSize(this.width, this.height)
            this.camera.updateProjectionMatrix()
            this.effectFXAA.uniforms['resolution'].value.set(
                1 / this.width,
                1 / this.height
            )
        },
        animate() {
            this.animationId = requestAnimationFrame(this.animate)
            this.composer.render(this.scene, this.camera)
            // this.renderer.render(this.scene, this.camera)

            // 根据相机控件可视化调试相机位置
            // console.log('controls.target', this.controls.target)
            // console.log('camera.position', this.camera.position)
        },
        showData() {
            console.log('controls.target', this.controls.target)
            console.log('camera.position', this.camera.position)
        },
        removeEventListeners() {},

        destroyThreejs() {
            try {
                this.renderer.dispose()
                this.renderer.forceContextLoss()
                this.renderer.content = null
                let gl = this.renderer.domElement.getContext('webgl')
                if (gl && gl.getExtension('WEBGL_lose_context')) {
                    gl.getExtension('WEBGL_lose_context').loseContext()
                }
                this.renderer = null
                this.camera = null
                this.scene.traverse((child) => {
                    if (child.material) {
                        child.material.dispose()
                    }
                    if (child.geometry) {
                        child.geometry.dispose()
                    }
                    child = null
                })
                this.scene = null
            } catch (e) {
                console.error('Failed to destroy threejs', e)
            }
        },
    },
    beforeDestroy() {
        cancelAnimationFrame(this.animationId)
        this.removeEventListeners()
        this.destroyThreejs()
        this.gui.domElement.style.display = 'none'
    },
}
</script>
<style lang='scss' scoped>
.model-container{
    position: relative;
    width: 100%;
    height: 100%;
    overflow: hidden;
    .view-container {
        width: 100%;
        height: 100%;
        overflow: hidden;
    }
    .tag-container{
        width: 240px;
        padding: 20px;
        border: 2px solid #fff;
        top: -80px;
        background-color: rgba(116, 130, 143, 0.9);
        color: #fff;
        border-radius: 10px;
        // box-shadow: 0 0 10px 4px #12f4e2;
    }
    #close{
        position: absolute;
        right: 10px;
        top: 10px;
        cursor: pointer;
    }
     #close:hover{
        color: #12f4e2;
    }
    .tag-container:after{
        content: '';
        height: 20px;
        width: 2px;
        background-color: #12f4e2;
        position: absolute;
        top: 100%;
        left: calc(50% - 1px);
    }
    .tag-container:before{
        content: '';
        height: 10px;
        width: 10px;
        border-radius: 5px;
        background-color: #12f4e2;
        position: absolute;
        top: calc(100% + 20px);
        left: calc(50% - 5px);
    }
    #loading{
        position: absolute;
        left: 0;
        top: 0;
        width: 100%;
        height: 100%;
        overflow: hidden;
        color: #fff;
        background-color: rgba(40, 85, 128, 0.9);
        display: flex;
        justify-content: center;
        align-items: center;
    }
    .btn{
        position: absolute;
        left: 50%;
        bottom: 20px;
        transform: translateX(-50%);
        z-index: 999;
    }
}

</style>